use chrono::{DateTime, Utc};
use serde::{Deserialize, Serialize};
use std::{fs::File, io::Write, path::Path};

#[derive(Serialize, Deserialize, Debug)]
pub struct HighscoreEntry {
    pub player_name: String,
    pub experience: usize,
    pub gold: usize,
    pub timestamp: i64,
}

impl HighscoreEntry {
    pub fn new(player_name: String, experience: usize, gold: usize) -> Self {
        HighscoreEntry {
            player_name,
            experience,
            gold,
            timestamp: Utc::now().timestamp(),
        }
    }
    pub fn get_date(&self) -> String {
        let dt = DateTime::from_timestamp(self.timestamp, 0).unwrap();
        format!("{}", dt.format("%Y-%m-%d"))
    }
}

pub type HighscoreList = Vec<HighscoreEntry>;

pub trait Sortable {
    fn insert_and_load<P: AsRef<Path>>(path: P, score: HighscoreEntry) -> Self;
}

impl Sortable for HighscoreList {
    fn insert_and_load<P: AsRef<Path>>(path: P, score: HighscoreEntry) -> Self {
        let file = File::open(&path);

        let mut result: HighscoreList = match file {
            Ok(file) => {
                if let Ok(result) = serde_json::from_reader(&file) {
                    result
                } else {
                    HighscoreList::new()
                }
            }
            Err(_) => HighscoreList::new(),
        };
        result.push(score);
        result.sort_by_key(|k| k.experience);
        result.reverse();
        let mut file = File::create(path).unwrap();
        file.write_all(serde_json::to_string(&result).unwrap().as_bytes())
            .unwrap();
        file.sync_all().unwrap();
        result
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    use tempfile::NamedTempFile;

    #[test]
    fn test_highscore_list() {
        let file = NamedTempFile::new().unwrap();
        let tmpfile = file.path();

        let list = HighscoreList::insert_and_load(
            Path::new(tmpfile),
            HighscoreEntry::new("noob".to_string(), 23, 0),
        );
        assert_eq!(list.len(), 1);
        assert_eq!(list.first().unwrap().experience, 23);

        let list = HighscoreList::insert_and_load(
            Path::new(tmpfile),
            HighscoreEntry::new("noob".to_string(), 42, 1),
        );
        assert_eq!(list.len(), 2);
        assert_eq!(list.first().unwrap().experience, 42);

        let list = HighscoreList::insert_and_load(
            Path::new(tmpfile),
            HighscoreEntry::new("noob".to_string(), 41, 2),
        );
        assert_eq!(list.len(), 3);
        assert_eq!(list.first().unwrap().experience, 42);
        assert_eq!(list[1].experience, 41);
    }
}
